"""
This submodule contains different tools to compute statistical
funtions for trajectory analysis and visualize the results.

All the resources of this module should be imported directly
from ``yupi.analyzing``. However, we ordered the resources
according the functionality into the following categories.
"""

from yupi.analyzing.processing import (
    turning_angles
)

from yupi.analyzing.transformations import (
    add_dynamic_reference,
    subsample_trajectory,
    wrap_theta
)

from yupi.analyzing.statistics import (
    estimate_turning_angles,
    estimate_velocity_samples,
    estimate_msd_ensemble,
    estimate_msd_time,
    estimate_msd,
    estimate_vacf_ensemble,
    estimate_vacf_time,
    estimate_vacf,
    estimate_kurtosis_ensemble,
    estimate_kurtosis_time,
    estimate_kurtosis
)

from yupi.analyzing.visualization import (
    plot_trajectories,
    plot_trajectory,
    plot_velocity_hist,
    plot_angle_distribution,
    plot_kurtosis,
    plot_msd,
    plot_vacf,
    LINE,
    DOTTED,
    LINE_DOTTED
)

__all__ = [
    'turning_angles',
    'add_dynamic_reference',
    'plot_trajectories',
    'plot_trajectory',
    'plot_velocity_hist',
    'plot_angle_distribution',
    'plot_kurtosis',
    'plot_msd',
    'plot_vacf',
    'LINE',
    'DOTTED',
    'LINE_DOTTED',
    'subsample_trajectory',
    'wrap_theta',
    'estimate_turning_angles',
    'estimate_velocity_samples',
    'estimate_msd_ensemble',
    'estimate_msd_time',
    'estimate_msd',
    'estimate_vacf_ensemble',
    'estimate_vacf_time',
    'estimate_vacf',
    'estimate_kurtosis_ensemble',
    'estimate_kurtosis_time',
    'estimate_kurtosis'
]



from typing import Tuple
import numpy as np
from yupi import Trajectory
from yupi.affine_estimator import affine_matrix


def add_dynamic_reference(traj: Trajectory,
                          reference: Tuple[np.ndarray, np.ndarray, np.ndarray],
                          start_at_origin=True):
    """
    This function fuses the information of a trajectory with an
    external reference of the motion of the Frame of Reference
    (FoR).

    It allows to remap the information gathered in local SoRs
    to a more general FoR.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    reference : Tuple[np.ndarray,np.ndarray,np.ndarray]
        Angular and translational parameters of the form
        ``(ang:np.ndarray, tx:np.ndarray, ty:np.ndarray)`` that
        accounts for the orientation and displacement of the reference.
    start_at_origin : bool, optional
        If True, set initial position at the origin. By default True.

    Returns
    -------
    Trajectory
        Output trajectory in the lab frame of reference.
    """

    def affine2camera(theta, tx, ty):
        x_cl, y_cl, theta_cl = np.zeros((3, theta.size + 1))
        theta_cl[1:] = np.cumsum(theta)

        for i in range(theta.size):
            A = affine_matrix(theta_cl[i + 1], x_cl[i], y_cl[i], R_inv=True)
            x_cl[i + 1], y_cl[i + 1] = A @ [-tx[i], -ty[i], 1]

        x_cl, y_cl, theta_cl = x_cl[1:], y_cl[1:], theta_cl[1:]
        return x_cl, y_cl, theta_cl

    def camera2obj(x_ac, y_ac, x_cl, y_cl, theta_cl):
        x_al, y_al = np.empty((2, x_ac.size))

        for i in range(x_ac.size):
            A = affine_matrix(theta_cl[i], x_cl[i], y_cl[i], R_inv=True)
            x_al[i], y_al[i] = A @ [x_ac[i], y_ac[i], 1]

        return x_al, y_al

    def affine2obj(theta, tx, ty, x_ac, y_ac):
        x_cl, y_cl, theta_cl = affine2camera(theta, tx, ty)
        x_al, y_al = camera2obj(x_ac, y_ac, x_cl, y_cl, theta_cl)
        return x_al, y_al

    theta, tx, ty = reference

    x_al, y_al = affine2obj(theta, tx, ty, traj.r.x, traj.r.y)

    if start_at_origin:
        x_al = x_al - x_al[0]
        y_al = y_al - y_al[0]

    traj.x = x_al
    traj.y = y_al

    return Trajectory(x=x_al, y=y_al, ang=traj.ang, t=traj.t, traj_id=traj.id)


def subsample_trajectory(traj: Trajectory, step=1, step_in_seconds=False):
    """
    Sample the trajectory ``traj`` by removing evenly spaced
    points according to ``step``.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    step : int, optional
        Number of sample points or period, depending on the value
        of ``step_in_seconds``. By default 1.
    step_in_seconds : bool, optional
        If True, ``step`` is considered as the number of sample
        points. Otherwise, ``step`` is interpreted as the sample
        period, in seconds. By default False.

    Returns
    -------
    Trajectory
        Output trajectory.
    """

    if step_in_seconds:
        step = int(step / traj.dt)

    points = traj.r[::step]
    ang = traj.ang[::step] if traj.ang is not None else None
    t = traj.t[::step] if traj.t is not None else None
    return Trajectory(points=points, t=t, ang=ang, dt=step*traj.dt)


def wrap_theta(ang: np.ndarray, degrees=False, centered=False):
    """
    Wrap angles by removing more than one lap to the
    trigonometic circle.

    Parameters
    ----------
    ang : np.ndarray
        Input array of angles.
    degrees : bool, optional
        If True, angles are given in degrees. Otherwise, the units
        are radians. By default False.
    centered : bool, optional
        If True, angles are wrapped on the interval ``[-pi, pi]``.
        Otherwise, the interval ``[0, 2*pi]`` is chosen. By default
        False.

    Returns
    -------
    np.ndarray
        A wrapped copy of ``ang``.
    """

    discont = 360 if degrees else 2 * np.pi
    if not centered:
        return ang % discont

    discont_half = discont / 2
    return -((discont_half - ang) % discont - discont_half)



import numpy as np
from yupi.trajectory import Trajectory
from yupi.analyzing.transformations import wrap_theta


def turning_angles(traj: Trajectory, accumulate=False, degrees=False,
                   centered=False):
    """
    Return the sequence of turning angles that forms the trajectory.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    accumulate : bool, optional
        If True, turning angles are measured with respect to an axis
        defined by the initial velocity (i.e., angles between initial
        and current velocity). Otherwise, relative turning angles
        are computed (i.e., angles between succesive velocity
        vectors). By default False.
    degrees : bool, optional
        If True, angles are given in degrees. Otherwise, the units
        are radians. By default False.
    centered : bool, optional
        If True, angles are wrapped on the interval ``[-pi, pi]``.
        Otherwise, the interval ``[0, 2*pi]`` is chosen. By default
        False.

    Returns
    -------
    np.ndarray
        Turning angles where each position in the array correspond
        to a given time instant.
    """

    dx = traj.delta_r.x
    dy = traj.delta_r.y
    theta = np.arctan2(dy, dx)

    if not accumulate:
        theta = np.ediff1d(theta)  # Relative turning angles
    else:
        theta -= theta[0]          # Rumulative turning angles

    return wrap_theta(theta, degrees, centered)



"""
This submodule is designed to capture Trajectory objects from
a sequence of images. It contains different tools to extract
the trajectories based on the color or the motion of the object
being tracked.

All the resources of this module should be imported directly
from ``yupi.tracking``. However, we ordered the resources
according the functionality into the following categories.
"""

from yupi.tracking.trackers import (
    ROI,
    ObjectTracker,
    CameraTracker,
    TrackingScenario
)
from yupi.tracking.algorithms import (
    TrackingAlgorithm,
    IntensityMatching,
    ColorMatching,
    FrameDifferencing,
    BackgroundSubtraction,
    TemplateMatching,
    OpticalFlow,
    BackgroundEstimator
)
from yupi.tracking.undistorters import (
    Undistorter,
    ClassicUndistorter,
    RemapUndistorter
)

__all__ = [
    'ROI',
    'ObjectTracker',
    'CameraTracker',
    'TrackingScenario',
    'TrackingAlgorithm',
    'IntensityMatching',
    'FrameDifferencing',
    'ColorMatching',
    'BackgroundSubtraction',
    'TemplateMatching',
    'OpticalFlow',
    'BackgroundEstimator',
    'Undistorter',
    'ClassicUndistorter',
    'RemapUndistorter',
    'NoUndistorter'
]



"""
The API of yupi is divided into four modules:
 *  yupi: General classes defining concepts used by all the other modules
 *  yupi.analyzing: Statistical methods and visualization tools to analize trajectories
 *  yupi.generating: Models to generate trajectories with given statistical constrains
 *  yupi.tracking: Tools to extract trajectories from image sequences based on the color or the motion of the object
    being tracked

All the resources of the library should be imported directly
from one of the aforementioned modules. However, we ordered the resources
according the functionality into subcategories to simplify searching.

"""

import logging
from yupi.trajectory import Trajectory
from yupi.vector import Vector

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s] %(message)s',
    datefmt='%H:%M:%S'
)

__all__ = [
    'Trajectory',
    'Vector'
]

__version__ = '0.5.5'


import abc
import numpy as np
from yupi import Trajectory


class Generator(metaclass=abc.ABCMeta):
    """
    Abstract class to model a Trajectory Generator. Classes inheriting
    from this class should implement ``generate`` method.

    Parameters
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.

    Attributes
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.
    n  : int
        Number of samples on each Trajectory.
    """

    def __init__(self, T: float, dim: int = 1, N: int = 1, dt: float = 1.0):
        # Siulation parameters
        self.T = T            # Total time
        self.dim = dim        # Trajectory dimension
        self.N = N            # Number of trajectories
        self.dt = dt          # Time step of the simulation
        self.n = int(T / dt)  # Number of time steps

    @abc.abstractmethod
    def generate(self):
        """
        Abstract method that is implemented on inheriting classes.
        It should compute a list of ``N`` Trajectory objects with the
        given parameters using a method specific to the inheriting class.
        """


class LatticeRandomWalkGenerator(Generator):
    """
    Multidimensional Lattice Random Walk Generator.


    Parameters
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.
    actions : np.ndarray, optional
        Vector of actions the walker can take, by default None.
    actions_prob : np.ndarray, optional
        Probability of every action to be taken according to
        every axis, by default None.
    jump_len : np.ndarray, optional
        Length of every single jump of the walker, by default None.
    """

    def __init__(self, T: float, dim: int = 1, N: int = 1, dt: float = 1,
                 actions: np.ndarray = None,
                 actions_prob: np.ndarray = None,
                 jump_len: np.ndarray = None):

        super().__init__(T, dim, N, dt)

        # Dynamic variables
        self.t = np.arange(self.n) * dt      # Time array
        self.r = np.zeros((self.n, dim, N))  # Position array

        # Model parameters
        # Only right/left jumps, uniform probabilities for it
        # And equal length for all jumps are set as default
        # TODO: Check that the model parameters received have the
        # expected shape
        if actions is None:
            actions = np.array([1, -1])
        if actions_prob is None:
            actions_prob = np.tile([.5, .5], (dim, 1))
        if jump_len is None:
            jump_len = np.ones((dim, N))

        self.actions = actions
        self.actions_prob = actions_prob
        self.jump_len = jump_len

    # Compute vector position as a function of time for
    # All the walkers of the ensemble
    def _get_r(self):
        # Get movements for every space coordinates according
        # To the sample space of probabilities in self.actions_prob
        dr = [np.random.choice(self.actions, p=p, size=(self.n - 1, self.N))
              for p in self.actions_prob]

        # Set time/coordinates as the first/second axis
        dr = np.swapaxes(dr, 0, 1)

        # Scale displacements according to the jump length statistics
        dr = dr * self.jump_len

        # Integrate displacements to get position vectors
        self.r[1:] = np.cumsum(dr, axis=0)
        return self.r

    # Get position vectors and generate RandomWalk object
    def generate(self):
        # Get position vectors
        r = self._get_r()

        # Generate RandomWalk object
        trajs = []
        for i in range(self.N):
            points = r[:, :, i]
            trajs.append(Trajectory(points=points, dt=self.dt, t=self.t,
                                    traj_id=f"Random Walker {i + 1}"))
        return trajs


class LangevinGenerator(Generator):
    """
    Random Walk class from a multidimensional Langevin Equation.

    Parameters
    ----------
    T : float
        Total duration of each Trajectory.
    dim : int, optional
        Dimension of each Trajectory, by default 1.
    N : int, optional
        Number of trajectories, by default 1.
    dt : float, optional
        Time step of the Trajectory, by default 1.0.
    tau : float, optional
        Relaxation characteristic time, by default 1.
    noise_pdf : str, optional
        Statistical model for the noise. ``noise_pdf`` should be a
        distribution from ``np.random``. By default 'normal'.
    noise_scale : float, optional
        Scale parameter of the noise, by default 1.
    v0 : np.ndarray, optional
        Initial velocities, by default None.
    r0 : np.ndarray, optional
        Initial positions, by default None.
    """

    def __init__(self, T: float, dim: int = 1, N: int = 1, dt: float = 1,
                 tau: float = 1.,
                 noise_pdf: str = 'normal',
                 noise_scale: float = 1,
                 v0: np.ndarray = None, r0: np.ndarray = None):

        super().__init__(T, dim, N, dt)

        # Model parameters

        # Relaxation time
        self.tau = tau
        # Noise PDF
        self.noise_pdf = noise_pdf
        # Scale parameter
        self.noise_scale = noise_scale
        # Noise array that will be filled in get_noise method
        self.noise = None

        # Dynamic variables

        # Shape of the dynamic variables
        self.shape = (self.n, dim, N)
        # Time array
        self.t = np.linspace(0, T, num=self.n)
        # Position array
        self.r = np.empty(self.shape)
        # Velocity array
        self.v = np.empty(self.shape)

        # Initial conditions
        # TODO: Check that r0 have the rigth shape
        self.r[0] = np.zeros((dim, N)) if r0 is None else r0
        # TODO: Check that v0 have the rigth shape
        self.v[0] = np.zeros((dim, N)) if v0 is None else v0

        self.v_scale = 1
        self.r_scale = 1
        self.t_scale = 1

    # Set intrinsic reference parameters
    # TODO: Check if scales are compatibles
    def set_scale(self, v_scale=None, r_scale=None, t_scale=None):
        if v_scale:
            self.v_scale = v_scale
        if r_scale:
            self.r_scale = r_scale
        if t_scale:
            self.t_scale = t_scale

    # Fill noise array with custom noise properties
    def _get_noise(self):
        dist = getattr(np.random, self.noise_pdf)
        self.noise = dist(scale=self.noise_scale, size=self.shape)

    # Solve Langevin Equation using the numerical method of Euler-Maruyama
    def _solve_rv(self):
        for i in range(self.n - 1):
            # Solving for position
            self.r[i + 1] = self.r[i] + \
                            self.v[i] * self.dt

            # Solving for velocity
            self.v[i + 1] = self.v[i] + \
                            -np.dot(1 / self.tau, self.v[i]) * self.dt + \
                            self.noise[i] * np.sqrt(self.dt)

    # Simulate the process
    def _simulate(self):
        self._get_noise()  # Set the attribute self.noise
        self._solve_rv()   # Solve the Langevin equation

    # Generate yupi Trajectory objects
    def generate(self):
        self._simulate()

        self.r *= self.r_scale
        self.v *= self.v_scale
        self.t *= self.t_scale
        self.dt *= self.t_scale

        trajs = []
        for i in range(self.N):
            points = self.r[:, :, i]
            trajs.append(Trajectory(points=points, dt=self.dt, t=self.t,
                                    traj_id=f"LangevinSolution {i + 1}"))
        return trajs



from typing import List
import scipy.stats
import numpy as np
from yupi.trajectory import Trajectory
from yupi.analyzing import turning_angles, subsample_trajectory


def estimate_turning_angles(trajs: List[Trajectory], accumulate=False,
                            degrees=False, centered=False):
    """Return a concatenation of all the turning angles that forms
    a set of trajectories.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    accumulate : bool, optional
        If True, turning angles are measured with respect to an axis
        define by the initial velocity (i.e., angles between initial
        and current velocity). Otherwise, relative turning angles
        are computed (i.e., angles between succesive velocity vectors).
        By default False.
    degrees : bool, optional
        If True, angles are given in degrees. Otherwise, the units
        are radians. By default False.
    centered : bool, optional
        If True, angles are wrapped on the interval ``[-pi, pi]``.
        Otherwise, the interval ``[0, 2*pi]`` is chosen. By default
        False.

    Returns
    -------
    np.ndarray
        Concatenated array of turning angles for a list of Trajectory
        objects.
    """

    theta = [turning_angles(t, accumulate, degrees, centered) for t in trajs]
    return np.concatenate(theta)


def estimate_velocity_samples(trajs: List[Trajectory], step: int = 1):
    """
    Estimate speeds of the list of trajectories, ``trajs``,
    by computing displacements according to a certain sample
    frequency given by ``step``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    step : int
        Numer of sample points.

    Returns
    -------
    np.array
        Concatenated array of speeds.
    """

    trajs_ = [subsample_trajectory(traj, step) for traj in trajs]
    return np.concatenate([traj.v.norm for traj in trajs_])


def estimate_msd_ensemble(trajs: List[Trajectory]):
    """
    Compute the square displacements for every Trajectory object
    stored in ``trajs`` as the square of the current position vector
    that has been subtracted the initial position.

    Trajectories should have the same length.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Array of square displacements with shape ``(n, N)``, where
        ``n`` is the total number of time steps and ``N`` the number
        of trajectories.
    """

    msd = []
    for traj in trajs:
        # Position vectors
        r = traj.r

        # Square displacements
        r_2 = (r - r[0])**2            # Square coordinates
        r2 = np.sum(r_2, axis=1)       # Square distances
        msd.append(r2)                 # Append square distances

    # Transpose to have time/trials as first/second axis
    msd = np.transpose(msd)
    return msd


def estimate_msd_time(trajs: List[Trajectory], lag: int):
    """
    Estimate the mean square displacement for every Trajectory
    object stored in ``trajs`` as the average of the square of
    dispacement vectors as a function of the lag time.

    This is a convenience estimator specially when trajectories
    do not have equal lengths.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of mean square displacements with shape ``(lag, N)``,
        where ``N`` the number of trajectories.
    """

    msd = []
    for traj in trajs:
        # Position vectors
        r = traj.r

        # Compute msd for a single trajectory
        msd_ = np.empty(lag)
        for lag_ in range(1, lag + 1):
            # Lag displacement vectors
            dr = r[lag_:] - r[:-lag_]
            # Lag displacement
            dr2 = np.sum(dr**2, axis=1)
            # Averaging over a single realization
            msd_[lag_ - 1] = np.mean(dr2)

        # Append all square displacements
        msd.append(msd_)

    # Transpose to have time/trials as first/second axis
    msd = np.transpose(msd)
    return msd


def estimate_msd(trajs: List[Trajectory], time_avg=True, lag=None):
    """
    Estimate the mean square displacement of the list of Trajectory
    objects, ``trajs``, providing the options of averaging over the
    ensemble of realizations or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, mean square displacement is estimated averaging over
        time. Otherwise, an ensemble average will be performed and all
        Trajectory objects will have to have the same length. By default
        True.
    lag : None, optional
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time. By
        default None.

    Returns
    -------
    np.ndarray
        Array of mean square displacements.
    np.ndarray
        Array of standard deviations.
    """

    if not time_avg:
        msd = estimate_msd_ensemble(trajs)   # Ensemble average
    else:
        msd = estimate_msd_time(trajs, lag)  # Time average

    msd_mean = np.mean(msd, axis=1)  # Mean
    msd_std = np.std(msd, axis=1)    # Standard deviation
    return msd_mean, msd_std


def estimate_vacf_ensemble(trajs: List[Trajectory]):
    """
    Compute the pair-wise dot product between initial and current
    velocity vectors for every Trajectory object stored in ``trajs``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Array of velocity dot products with shape ``(n, N)``, where
        ``n`` is the total number of time steps and ``N`` the number
        of trajectories.
    """

    vacf = []
    for traj in trajs:
        # Cartesian velocity components
        v = traj.v

        # Pair-wise dot product between velocities at t0 and t
        v0_dot_v = np.sum(v[0] * v, axis=1)

        # Append all veloctiy dot products
        vacf.append(v0_dot_v)

    # Transpose to have time/trials as first/second axis
    vacf = np.transpose(vacf)
    return vacf


def estimate_vacf_time(trajs: List[Trajectory], lag: int):
    """
    Estimate the velocity autocorrelation function for every
    Trajectory object stored in ``trajs`` as the average of the
    dot product between velocity vectors that are distant a certain
    lag time.

    This is a convenience estimator specially when trajectories do
    not have equal lengths.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of velocity autocorrelation function with shape
        ``(lag, N)``, where ``N`` is the number of trajectories.
    """

    vacf = []
    for traj in trajs:
        # Cartesian velocity components
        v = traj.v

        # Compute vacf for a single trajectory
        vacf_ = np.empty(lag)
        for lag_ in range(1, lag + 1):
            # Multiply components given lag
            v1v2 = v[:-lag_] * v[lag_:]

            # Dot product for a given lag time
            v1_dot_v2 = np.sum(v1v2, axis=1)

            # Averaging over a single realization
            vacf_[lag_ - 1] = np.mean(v1_dot_v2)

        # Append the vacf for a every single realization
        vacf.append(vacf_)

    # Aranspose to have time/trials as first/second axis
    vacf = np.transpose(vacf)
    return vacf


def estimate_vacf(trajs: List[Trajectory], time_avg=True, lag: int = None):
    """
    Estimate the velocity autocorrelation function of the list of
    Trajectory objects, ``trajs``, providing the options of averaging
    over the ensemble of realizations or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, velocity autocorrelation function is estimated
        averaging over time. Otherwise, an ensemble average will be
        performed and all Trajectory objects will have to have the
        same length. By default True.
    lag : int, optional
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time.
        By default None.

    Returns
    -------
    np.ndarray
        Array of velocity autocorrelation function.
    np.ndarray
        Array of standard deviations.
    """

    if not time_avg:
        vacf = estimate_vacf_ensemble(trajs)   # Ensemble average
    else:
        vacf = estimate_vacf_time(trajs, lag)  # Time average

    vacf_mean = np.mean(vacf, axis=1)  # Mean
    vacf_std = np.std(vacf, axis=1)    # Standard deviation
    return vacf_mean, vacf_std


# kurtosis (ensemble average)
# TODO: Fix this implementation for dim != 2 Traj
def estimate_kurtosis_ensemble(trajs: List[Trajectory]):
    """[summary]

    Parameters
    ----------
    trajs : List[Trajectory]
        [description]

    Returns
    -------
    [type]
        [description]
    """

    kurtosis = []
    for traj in trajs:
        dx = traj.delta_r.x
        dy = traj.delta_r.y
        kurt = np.sqrt(dx**2 + dy**2)
        kurtosis.append(kurt)
    return scipy.stats.kurtosis(kurtosis, axis=0, fisher=False)


# kurtosis (time average)
# TODO: Fix this implementation for dim != 2 Traj
def estimate_kurtosis_time(trajs: List[Trajectory], lag):
    """[summary]

    Parameters
    ----------
    trajs : List[Trajectory]
        [description]
    lag : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """

    kurtosis = []
    for traj in trajs:
        kurt = np.empty(lag)
        for lag_ in range(1, lag + 1):
            dx = traj.r.x[lag_:] - traj.r.x[:-lag_]
            dy = traj.r.y[lag_:] - traj.r.y[:-lag_]
            dr = np.sqrt(dx**2 + dy**2)
            kurt[lag_ - 1] = scipy.stats.kurtosis(dr, fisher=False)
        kurtosis.append(kurt)
    return np.mean(kurtosis, axis=0)


# get displacements for ensemble average and
# kurtosis for time average
def estimate_kurtosis(trajs: List[Trajectory], time_avg=True, lag=None):
    """[summary]

    Parameters
    ----------
    trajs : List[Trajectory]
        [description]
    time_avg : bool, optional
        [description], by default True
    lag : [type], optional
        [description], by default None

    Returns
    -------
    [type]
        [description]
    """

    if not time_avg:
        return estimate_kurtosis_ensemble(trajs)
    return estimate_kurtosis_time(trajs, lag)



import logging
from typing import List
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from yupi.trajectory import Trajectory


LINE = '-'
DOTTED = 'o'
LINE_DOTTED = '-o'

# TODO: Fix this implementation for dim != 2
def plot_trajectories(trajs: List[Trajectory], line_style: str = LINE,
                      max_trajectories: bool = None, title: str = None,
                      legend: bool = True, show: bool = True,
                      connected: bool = False, unit: str = None, color = None,
                      **kwargs):
    """
    Plot all or ``max_trajectories`` trajectories from ```trajs``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input trajectories.
    line_style : str
        Type of the trajectory line to plot. It uses the matplotlib,
        notation, by default '-'.
    max_trajectories : int, optional
        Number of trajectories to plot, by default None.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If True, the plot is shown. By default True.
    connected : bool
        If True, all the trajectory points of same index will be,
        connected.

        If the trajectories do not have same length then the points
        will be connected until the shortest trajectory last index.
    color : str or tuple or list
        Defines the color of the trajectories, by default None.

        If color is of type ``str`` or ``tuple`` (rgb) then the color
        is applied to all trajectories. If color is of type ``list``
        then the trajectories take the color according to the index.

        If there are less colors than trajectories then the remaining
        trajectories are colored automatically (not with the same
        color).
    """

    if max_trajectories is None:
        max_trajectories = len(trajs)

    unit = '' if unit is None else f' [{unit}]'

    colors = None
    if color is not None:
        if isinstance(color, (str, tuple)):
            kwargs['color'] = color
        elif isinstance(color, list):
            colors = color

    if connected:
        lengths = list(map(len, trajs))
        min_len = min(lengths)
        max_len = max(lengths)
        if min_len != max_len:
            logging.warning('Not all the trajectories have the same length.')
        for i in range(min_len):
            traj_points = [t[i] for t in trajs]
            traj_points.append(traj_points[0])
            for tp1, tp2 in zip(traj_points[:-1], traj_points[1:]):
                xs = [tp1.r[0], tp2.r[0]]
                ys = [tp1.r[1], tp2.r[1]]
                plt.plot(xs, ys, color=(0.2, 0.2, 0.2), linewidth=0.5)

    for i, t in enumerate(trajs):
        if i == max_trajectories:
            break

        # Plotting
        x, y = t.r.x, t.r.y
        if colors is not None:
            if i < len(colors):
                kwargs['color'] = colors[i]
            else:
                kwargs.pop('color')
        traj_plot = plt.plot(x, y, line_style, **kwargs)
        color = traj_plot[-1].get_color()
        plt.plot(x[0], y[0], 'o', mfc='white', zorder=2,
                 label=f'{t.id} initial position', color=color)
        plt.plot(x[-1], y[-1], 'o', mfc='white', zorder=2, color=color)
        plt.plot(x[-1], y[-1], 'o', alpha=.5,
                 label=f'{t.id} final position', color=color)

        if legend:
            plt.legend()

        plt.title(title)
        plt.tick_params(direction='in')
        plt.axis('equal')
        plt.grid(True)
        plt.xlabel(f'x{unit}')
        plt.ylabel(f'y{unit}')

    if show:
        plt.show()


def plot_trajectory(traj: Trajectory, title=None, legend=True, show=True):
    """
    Plot a single trajectory.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If Tue, the plot is shown. By default True.
    """

    plot_trajectories([traj], title=title, legend=legend, show=show)


def plot_velocity_hist(v, bins=20, show=True):
    """[summary]

    Parameters
    ----------
    v : [type]
        [description]
    bins : int, optional
        [description], by default 20
    show : bool, optional
        [description], by default True
    """

    plt.hist(v, bins, density=True, ec='k', color='#fdd693')
    plt.xlabel('speed [m/s]')
    plt.ylabel('pdf')
    if show:
        plt.show()


def plot_angle_distribution(ang, bins=50, ax=None, show=True):
    """[summary]

    Parameters
    ----------
    ang : [type]
        [description]
    bins : int, optional
        [description], by default 50
    ax : [type], optional
        [description], by default None
    plot : bool, optional
        [description], by default True
    """

    if ax is None:
        ax = plt.gca(projection='polar')
    plt.hist(ang, bins, density=True, ec='k', color='.85')
    ax.set_theta_zero_location('N')
    ax.set_rlabel_position(135)
    ax.set_axisbelow(True)
    plt.xlabel('turning angles pdf')
    if show:
        plt.show()


def plot_msd(msd, msd_std, dt, lag=30, show=True):
    """[summary]

    Parameters
    ----------
    msd : [type]
        [description]
    msd_std : [type]
        [description]
    dt : [type]
        [description]
    lag : int, optional
        [description], by default 30
    show : bool, optional
        [description], by default True
    """

    lag_t_msd = dt * np.arange(lag)
    plt.plot(lag_t_msd, msd, color='.2')
    plt.fill_between(lag_t_msd, msd + msd_std, msd - msd_std, color='#afc0da')
    plt.xlabel('lag time [s]')
    plt.ylabel(r'$\mathrm{msd \; [m^2/s]}$')
    if show:
        plt.show()


def plot_kurtosis(kurtosis, dt=None, t_array=None, show=True):
    """[summary]

    Parameters
    ----------
    kurtosis : [type]
        [description]
    dt : [type], optional
        [description], by default None
    t_array : [type], optional
        [description], by default None
    plot : bool, optional
        [description], by default True
    """

    if dt:
        t_array = np.linspace(0, dt*len(kurtosis), len(kurtosis))
    if t_array is not None:
        plt.plot(t_array, kurtosis)
        plt.xlabel('time [s]')
    else:
        plt.plot(kurtosis)

    plt.ylabel('kurtosis')
    if show:
        plt.show()


def plot_vacf(vacf, dt, lag=50, show=True):
    """[summary]

    Parameters
    ----------
    vacf : [type]
        [description]
    dt : [type]
        [description]
    lag : int, optional
        [description], by default 50
    show : bool, optional
        [description], by default True
    """

    lag_t_vacf = dt * np.arange(lag)

    plt.plot(lag_t_vacf, vacf, '.', color='#870e11', mfc='w')
    plt.xlabel('lag time [s]')
    plt.ylabel(r'$\mathrm{vacf \; [(m/s)^2]}$')

    ax = plt.gca()

    inset_axes(ax, width='60%', height='60%', bbox_to_anchor=(0, 0, 1, 1),
               bbox_transform=ax.transAxes, loc='upper right')

    plt.plot(lag_t_vacf, vacf, '.', color='#870e11', mfc='w')
    plt.yscale('log')

    if show:
        plt.show()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yupi', 'yupi.analyzing', 'yupi.generating', 'yupi.tracking']

package_data = \
{'': ['*']}

install_requires = \
['PyQt5>=5.15.3,<6.0.0',
 'matplotlib>=3.2.0',
 'nudged>=0.3.1',
 'numpy>=1.16.5',
 'opencv-python>=4.4.0',
 'scipy>=1.6.1,<2.0.0']

setup_kwargs = {
    'name': 'yupi',
    'version': '0.5.5',
    'description': 'A package for tracking and analysing objects trajectories',
    'long_description': '# yupi\n\nStanding for *Yet Underused Path Instruments*, yupi is a set of tools designed \nfor collecting, generating and processing trajectory data.  \n\n## Installation\n\nCurrent recommended installation method is via the pypi package:\n\n```cmd\npip install yupi\n```\n\n## Getting Started\n\nIn the [official documentation](https://yupi.readthedocs.io/en/latest/) there \nare some resources to start using the library. A [Getting Started section](https://yupi.readthedocs.io/en/latest/getting_started/getting_started.html) provides an introduction for\nnewcomers and the [API reference section](https://yupi.readthedocs.io/en/latest/api_reference/api_reference.html) contains a detailed description of the API.\n\n\n## Examples\n\nCode examples (with additional multimedia resources) can be found in \n[this repository](https://github.com/yupidevs/yupi_examples). Additionally, in\nthe [Examples section](https://yupi.readthedocs.io/en/latest/examples/examples.html)\nof the documentation, you can find the same examples with additional comments \nand expected execution results in order to inspect the examples without actually \nexecuting them.',
    'author': 'Gustavo Viera-López',
    'author_email': 'gvieralopez@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
